@@ -1,23 +1,22 @@
 #
 # By Krister Hedfors
 #
+# NOTE: custom valgetter BREAKS stuff
+# as __getitem__ is used by the other dict methods as well
+# All key/val modifiers can in fact break stuff severely.
+#
 # + pack/unpack everywhere
 # + db-sorted iteritems
 # |- inform about broken ordering of anything but strings
 # + ns functionality transparently optional
 # + create class with a mapping of an existing table
 # + move ns- and flags stuff to dedicated functions
+# + getrow returns row and valcol is not required then
+#
+# - fix dup()
+# - fix keys(), values() etc for custom_queries, raise or something
+# - optional uniqueness
 #
-# - add user controlled key preprocessor and val processor
-# - add raw query template as valgetter
-# - if valgetter is a string, it's used as an SQL template
-#       which is fed with key to obtain val
-# - build bunch of rawsql-dicts and build code snipplets
-#       with which to extract values
-# - build programs using these rawsql-dicts as components
-# - joins work just as well, as it's raw dicts
-# - iterations simply leave out which constraint
-# keymod, valmod, valmodin alongside valgetter
 #
 #
 __all__ = ['sqldict']
@@ -29,6 +28,9 @@ __doc__ = '''
 
 
 import doctest
+import pydoc
+
+import sys
 from UserDict import DictMixin
 import pickle
 import cPickle
@@ -53,23 +55,20 @@ _KEYSIZE = 1000-64
 _VALSIZE = 4000
 _NSSIZE  = 64
 
-_DEFAULT_KEYCOL   = "sqldict_key"
-_DEFAULT_VALCOL   = "sqldict_val"
-_DEFAULT_NSCOL    = "sqldict_ns"
-_DEFAULT_FLAGSCOL = "sqldict_flags"
-_DEFAULT_VALGETTER = lambda row:row[_DEFAULT_VALCOL]
-_DEFAULT_KEYMOD   = lambda key:key
-_DEFAULT_VALMODIN = lambda val:val
-_DEFAULT_VALMODOUT = lambda val:val
+_DEFAULT_ECHO      = None # or file
+_DEFAULT_KEYCOL    = "sqldict_key"
+_DEFAULT_VALCOL    = "sqldict_val"
+_DEFAULT_NSCOL     = "sqldict_ns"
+_DEFAULT_FLAGSCOL  = "sqldict_flags"
+_DEFAULT_VALGETTER = lambda row : row[_DEFAULT_VALCOL]
+_DEFAULT_KEYMOD    = lambda key : key
+_DEFAULT_VALMODIN  = lambda val : val
+_DEFAULT_VALMODOUT = lambda val : val
 
 _DEFAULT_KEYTYPE=String(_KEYSIZE)
 _DEFAULT_VALTYPE=String(_VALSIZE)
 
-
-def log_execute(conn, query):
-    #print str(query)
-    result = conn.execute(query)
-    return result
+_DEFAULT_COLUMNS = []
 
 
 class SetConfig(object):
@@ -91,7 +90,19 @@ class SQLDict(DictMixin, SetConfig):
     _F_UNICODE_VAL    = 1 << 3
 
 
+    def log_execute(self, conn, query):
+        echo = self._echo
+        if echo:
+            if hasattr(echo, "write"):
+                echo.write(str(query))
+            else:
+                print query
+        result = conn.execute(query)
+        return result
+
+
     def __init__(self,  conn, tablename,
+                            echo=_DEFAULT_ECHO,
                             create=0,
                             ns='',
                             keycol=_DEFAULT_KEYCOL,
@@ -101,13 +112,15 @@ class SQLDict(DictMixin, SetConfig):
                             keytype=_DEFAULT_KEYTYPE,
                             valtype=_DEFAULT_KEYTYPE,
                             valgetter=_DEFAULT_VALGETTER,
+                            getrow=0,
                             keymod=_DEFAULT_KEYMOD,
                             valmodin=_DEFAULT_VALMODIN,
                             valmodout=_DEFAULT_VALMODOUT,
+                            query=None,
                             extended=0, # 0 or 1; simple or complex types
                             sort=1,         # 0=dont, 1=keys, 2=values(broken)
                             iterate=0,
-
+                            columns=_DEFAULT_COLUMNS,
                             nssize=_NSSIZE,
                             ):
         '''
@@ -116,6 +129,7 @@ class SQLDict(DictMixin, SetConfig):
             self._conn = conn.connect()
         else:
             self._conn = conn
+        self._echo = echo
         self._tablename = tablename
         self._keycol = keycol #as string name
         self._valcol = valcol #as string name
@@ -124,14 +138,17 @@ class SQLDict(DictMixin, SetConfig):
         self._keytype = keytype
         self._valtype = valtype
         self._valgetter = valgetter
+        self._getrow = getrow
         self._keymod = keymod
         self._valmodin = valmodin
         self._valmodout = valmodout
+        self._custom_query = query
         self._extended = extended
         self._ns = ns
         self._create = create
         self._sort = sort
         self._iterate = iterate
+        self._columns = columns
         self._nssize =  nssize
         self._meta = MetaData()
         self._meta.bind = conn
@@ -143,25 +160,52 @@ class SQLDict(DictMixin, SetConfig):
         # If user changes valcol but not valgetter we need
         # to set it for him, otherwise __getitem__ wont work.
         #
-        if valcol != _DEFAULT_VALCOL:
-            if valgetter != _DEFAULT_VALGETTER:
-                pass
+        if getrow:
+            if valgetter == _DEFAULT_VALGETTER:
+                valgetter = lambda row : row
             else:
-                self._valgetter = lambda row:row[valcol]
+                msg = "valgetter and getrow cant be used together"
+                raise SQLDictError(msg)
+
+        elif self._custom_query and valgetter == _DEFAULT_VALGETTER:
+            valgetter = lambda row:row[0]
+
+        elif valcol != _DEFAULT_VALCOL and valgetter == _DEFAULT_VALGETTER:
+            valgetter = lambda row:row[valcol]
+
+        self._valgetter = valgetter
+
+        if columns and not create:
+            msg = "columns has no effect unless `create=1`"
+            raise SQLDictError(msg)
 
         #if create:
         #    if self._extended is None:
         #        self._extended = 1
-        try:
-            self._table = self._autoload_table()
-        except NoSuchTableError, e:
-            if create:
-                self._create_table()
-            else:
+        if create:
+            self._create_table(columns)
+        else:
+            try:
+                self._table = self._autoload_table()
+            except NoSuchTableError, e:
                 msg = 'initialize sqldict with `create=1` to create table'
                 error = (msg, e)
                 raise NoSuchTableError(error)
 
+        #try:
+        #    self._table = self._autoload_table()
+        #except NoSuchTableError, e:
+        #    if create:
+        #        self._create_table(columns)
+        #    else:
+        #        msg = 'initialize sqldict with `create=1` to create table'
+        #        error = (msg, e)
+        #        raise NoSuchTableError(error)
+
+    def _render_query_from_template(self, key):
+        query = self._custom_query % key
+        return query
+
     def _autoload_table(self):
         tablename = self._tablename
         meta = self._meta
@@ -173,7 +217,7 @@ class SQLDict(DictMixin, SetConfig):
         self._keyvaltype = type( getattr(table.c, keyval))
         return table
 
-    def _create_table(self):
+    def _create_table(self, columns=[]):
         keycol = self._keycol
         valcol = self._valcol
         nscol = self._nscol
@@ -184,18 +228,19 @@ class SQLDict(DictMixin, SetConfig):
         ns = self._ns
         extended = self._extended
         #
-        columns = [
-            Column(keycol, keytype, nullable=False, primary_key=True),
-            Column(valcol, valtype, nullable=False),
-        ]
-        if extended:
-            columns.append(
-                Column(flagscol, Integer, nullable=False),
-            )
-        if ns:
-            columns.append(
-                Column(nscol,  String(nssize), nullable=False, primary_key=True)
-            )
+        if not columns:
+            columns = [
+                Column(keycol, keytype, nullable=False, primary_key=True),
+                Column(valcol, valtype, nullable=False),
+            ]
+            if extended:
+                columns.append(
+                    Column(flagscol, Integer, nullable=False),
+                )
+            if ns:
+                columns.append(
+                    Column(nscol,  String(nssize), nullable=False, primary_key=True)
+                )
         table = Table(self._tablename, self._meta, *columns)
         self._table = table
         self._meta.create_all()
@@ -210,12 +255,10 @@ class SQLDict(DictMixin, SetConfig):
         obj = pickle.loads(blob)
         return obj
 
-    def _update(self, key, val, flags):
-        'key and or val must be serialized according to flags'
+    def _update_values(self, key, values, flags):
         table = self._table
         conn = self._conn
         keycol = self._keycol
-        valcol = self._valcol
         flagscol = self._flagscol
         extended = self._extended
         key_column = self._get_key_column()
@@ -224,15 +267,20 @@ class SQLDict(DictMixin, SetConfig):
         query = self._apply_ns_where_clause(query)
         query = query.where(key_column == key)
 
-        values = {
-            valcol : val
-        }
-
         if extended:
             values[flagscol] = flags
 
         query = query.values(**values)
-        log_execute(conn, query)
+        result = self.log_execute(conn, query)
+        return result.rowcount
+
+    def _update(self, key, val, flags):
+        'key and or val must be serialized according to flags'
+        valcol = self._valcol
+        values = {
+            valcol : val
+        }
+        return self._update_values(key, values, flags)
 
     def drop(self):
         self._table.drop()
@@ -345,39 +393,44 @@ class SQLDict(DictMixin, SetConfig):
         #
         if extended:
             columns += [self._get_flags_column()]
+        if self._getrow:
+            columns = list(table.c)
         query = select(columns)
         query = self._apply_ns_where_clause(query)
         query = self._apply_ordering(query)
         return query
 
     def _select_items(self):
-        key_column = self._get_key_column()
-        val_column = self._get_val_column()
+        getrow = self._getrow
         #
-        columns = [
-            key_column,
-            val_column,
-        ]
+        columns = []
+        if not getrow:
+            columns = [
+                self._get_key_column(),
+                self._get_val_column()
+            ]
         return self._select(columns)
 
     def _select_keys(self):
-        key_column = self._get_key_column()
-        val_column = self._get_val_column()
+        getrow = self._getrow
         #
-        columns = [
-            key_column,
-            val_column,
-        ]
+        columns = []
+        if not getrow:
+            columns = [
+                self._get_key_column(),
+                self._get_val_column()
+            ]
         return self._select(columns)
 
     def _select_values(self):
-        key_column = self._get_key_column()
-        val_column = self._get_val_column()
+        getrow = self._getrow
         #
-        columns = [
-            key_column,
-            val_column,
-        ]
+        columns = []
+        if not getrow:
+            columns = [
+                self._get_key_column(),
+                self._get_val_column()
+            ]
         return self._select(columns)
 
 
@@ -393,7 +446,7 @@ class SQLDict(DictMixin, SetConfig):
         #
         query = self._select_items()
         #print query
-        result = log_execute(conn, query)
+        result = self.log_execute(conn, query)
         for row in result:
             key = row[keycol]
             val = valgetter(row)
@@ -411,7 +464,7 @@ class SQLDict(DictMixin, SetConfig):
         extended = self._extended
         #
         query = self._select_keys()
-        result = log_execute(conn, query)
+        result = self.log_execute(conn, query)
         for row in result:
             key = row[keycol]
             flags = 0
@@ -427,7 +480,7 @@ class SQLDict(DictMixin, SetConfig):
         valgetter = self._valgetter
         #
         query = self._select_values()
-        result = log_execute(conn, query)
+        result = self.log_execute(conn, query)
         for row in result:
             val = valgetter(row)
             flags = 0
@@ -454,7 +507,7 @@ class SQLDict(DictMixin, SetConfig):
         if sort == 2:
             query = query.order_by(ns_column)
         #print query
-        result = log_execute(conn, query)
+        result = self.log_execute(conn, query)
         namespaces = []
         for row in result:
             ns = row[nscol]
@@ -486,23 +539,28 @@ class SQLDict(DictMixin, SetConfig):
         return duplicate
 
     def keys(self):
-        conn = self._conn
-        extended = self._extended
-        keycol = self._keycol
-        flagscol = self._flagscol
-        #
-        query = self._select_keys()
-        #print query
-        result = log_execute(conn, query)
-        keys = []
-        for row in result:
-            key = row[keycol]
-            flags = 0
-            if extended:
-                flags = row[flagscol]
-            key = self._unpack_key(key, flags)
-            keys.append(key)
-        return keys
+        return list(self.iterkeys())
+    #def keys(self):
+    #    conn = self._conn
+    #    extended = self._extended
+    #    keycol = self._keycol
+    #    flagscol = self._flagscol
+    #    #
+    #    query = self._select_keys()
+    #    #print query
+    #    result = self.log_execute(conn, query)
+    #    keys = []
+    #    for row in result:
+    #        key = row[keycol]
+    #        flags = 0
+    #        if extended:
+    #            flags = row[flagscol]
+    #        key = self._unpack_key(key, flags)
+    #        keys.append(key)
+    #    return keys
+
+    def iterate(self, key):
+        return self.__iter_getitem__(key, ordering=1)
 
     def __iter_getitem__(self, key, ordering=0):
         #print 'getitem', key
@@ -516,20 +574,26 @@ class SQLDict(DictMixin, SetConfig):
         valgetter = self._valgetter
         keymod = self._keymod
         valmodout = self._valmodout
+        custom_query = self._custom_query
         #
         if keymod:
             key = keymod(key)
         key, _ = self._pack_key(key, 0)
-        query = select([table])
-        query = self._apply_ns_where_clause(query)
-        query = query.where( key_column == key)
-        if ordering:
-            query = self._apply_ordering(query)
 
-        result = log_execute(conn, query)
+        if custom_query:
+            query = self._render_query_from_template(key)
+        else:
+            query = select([table])
+            query = self._apply_ns_where_clause(query)
+            query = query.where( key_column == key)
+            if ordering:
+                query = self._apply_ordering(query)
+
+        result = self.log_execute(conn, query)
         for row in result:
             #
             val = valgetter(row)
+
             flags = 0
             if extended:
                 flags = row[flagscol]
@@ -562,7 +626,7 @@ class SQLDict(DictMixin, SetConfig):
         query = self._apply_ns_where_clause(query)
         query = self._apply_key_where_clause(query, key)
 
-        result = log_execute(conn, query)
+        result = self.log_execute(conn, query)
         if result.rowcount == 0:
             raise KeyError
         elif result.rowcount > 1:
@@ -570,9 +634,8 @@ class SQLDict(DictMixin, SetConfig):
             raise SQLDictException(error)
         return
 
-    def __setitem__(self, key, val):
+    def setvalues(self, key, values={}, **kwvalues):
         keycol = self._keycol
-        valcol = self._valcol
         flagscol = self._flagscol
         nscol = self._nscol
         extended = self._extended
@@ -580,32 +643,71 @@ class SQLDict(DictMixin, SetConfig):
         keymod = self._keymod
         valmodin = self._valmodin
         #
+        values.update(kwvalues)
+
+        if keymod:
+            key = keymod(key)
+        #if valmodin:
+        #    val = valmodin(val)
+
+        key, flags = self._pack_key(key, 0)
+        #val, flags = self._pack_val(val, flags)
+
+        rowcount = self._update_values(key, values, flags)
+        if rowcount == 0:
+            values.setdefault(keycol, key)
+            self._insert_values(key, values, flags)
+
+
+    def __setitem__(self, key, val):
+        keycol = self._keycol
+        valcol = self._valcol
+        keymod = self._keymod
+        valmodin = self._valmodin
+        extended = self._extended
+        #
         if keymod:
             key = keymod(key)
         if valmodin:
             val = valmodin(val)
 
         key, flags = self._pack_key(key, 0)
-        val, flags = self._pack_val(val, flags)
+        if type(val) is dict and not extended:
+            values = val
+            values.setdefault(keycol, key)
+            rowcount = self._update_values(key, values, flags)
+            if rowcount == 0:
+                self._insert_values(key, values, flags)
+        else:
+            val, flags = self._pack_val(val, flags)
+            rowcount = self._update(key, val, flags)
+            if rowcount == 0:
+                values = {
+                    keycol : key,
+                    valcol : val,
+                }
+                self._insert_values(key, values, flags)
 
+    def _insert_values(self, key, values, flags):
+        keycol = self._keycol
+        valcol = self._valcol
+        extended = self._extended
         conn = self._conn
         table = self._table
+        flagscol = self._flagscol
+        ns = self._ns
+        nscol = self._nscol
+        #
         query = insert(table)
-
-        values = {
-            keycol : key,
-            valcol : val,
-        }
+        values.setdefault(keycol, key)
         if extended:
             values[flagscol] = flags
         if ns:
             values[nscol] = ns
 
         query = query.values(**values)
-        try:
-            log_execute(conn, query)
-        except IntegrityError:
-            self._update(key, val, flags)
+        result = self.log_execute(conn, query)
+        return result
 
 
 
@@ -664,37 +766,84 @@ def sqldict(db, tablename,
                     keytype=_DEFAULT_KEYTYPE,
                     valtype=_DEFAULT_VALTYPE,
                     valgetter=_DEFAULT_VALGETTER,
+                    getrow=0,
                     keymod=_DEFAULT_KEYMOD,
                     valmodin=_DEFAULT_VALMODIN,
                     valmodout=_DEFAULT_VALMODOUT,
+                        echo=_DEFAULT_ECHO,
+                        query=None,
                         iterate=0,
                         serialize=0,
                         create=0,
                         sort=1,
                         ns=None,
+                        columns=_DEFAULT_COLUMNS,
                         ):
 
     """
-    The various dictionaries, dN, can be used just as an ordinary dict.
+    Welcome, please read through this brief documentation.
+
+    All the sqldict examples below will use an sqlalchemy engine e.
+    Certain database engines will need certain back-end modules
+    which in most cases are available as packages in your *nix
+    distribution or as .exe installers, see INSTALLATION* below.
+
+    Lets create the engine object:
+
+    >>> from sqlalchemy import *
+    >>> dburi = 'mysql://user:pass@localhost/sqldict'
+    >>> e = create_engine(dburi)
+
+    Create database tables, t1 and t2, to hold dicts d1 and d2:
+
+    >>> d1 = sqldict(e, "t1", create=1)
+    >>> d2 = sqldict(e, "t2", create=1, keytype=String(50), valtype=Integer)
+
+    An sqldict() support all or most common dict operations depending on
+    which sqldict extensions are chosen.
+
+    This is a selection of some useful sqldict instantiation args:
+
+        echo          - print all executed sql queries to stdout (bool)
+        create        - create database table (bool)
+        sort          - 0=no sort, 1=keys, 2=values
+        serialize     - transparent pickling for non-String type values
+                          assigned to String-type columns (bool)
+        getrow        - get entire row instead of just val (bool)
+        keycol        - name of key column
+        valcol        - name of val column
+        keytype       - sqlalchemy column type for key
+        valtype       - sqlalchemy column type for val
+
+    Actions taken such as setting/getting items result in logically
+    corresponding SQL statements. Some examples are:
+
+        __getitem__() - SELECT keycol FROM tablename WHERE valcol=%(val)s
+        iter*()       - iterate over result-set from a select statement
+        keys|values() - returns list(iter*())
+        __del__()     - DELETE FROM..
+        __setitem__() - UPDATE.., INSERT..
+
+
 
-    >>> #d0 = sqldict("mysql://user:pass@dbhost/dbname", "table0", create=1)
-    >>> e = create_engine("sqlite://")
-    >>> d1 = sqldict(e, "table2", create=1)
-    >>> d2 = sqldict(e, "table3", create=2, valtype=Integer)
     >>> contents = {"asd":123}
     >>> d1.update(contents)
     >>> d2.update(contents)
     >>> assert d1["asd"] == "123"
     >>> assert d2["asd"] == 123
+    >>> d1.drop()
+    >>> d2.drop()
 
-    >>> _ = e.execute("create table asd (i integer, s varchar(50))")
-    >>> _ = e.execute("insert into asd values (42, 'gubbe')")
-    >>> _ = e.execute("insert into asd values (99, 'hatt')")
+    >>> assert e.execute("create table asd (i integer, s varchar(50))")
+    >>> assert e.execute("insert into asd values (42, 'gubbe')")
+    >>> assert e.execute("insert into asd values (99, 'hatt')")
     >>> d3 = sqldict(e, "asd", keycol="s", valcol="i")
     >>> d4 = sqldict(e, "asd", keycol="i", valcol="s")
     >>> #d5 = sqldict(e, "asd", keycol="s", valgetter=lambda r:r)
     >>> assert d3["gubbe"] == 42
     >>> assert d4[42] == "gubbe"
+    >>> assert e.execute("drop table asd")
+    >>>
     >>> #assert d5["gubbe"]["i"] == 42
 
     broken (at least mysql) when combining keytype and serialize,
@@ -703,6 +852,13 @@ def sqldict(db, tablename,
     >>> from sqlalchemy import *
     >>> d = sqldict(e, "laarge", create=1, serialize=1, keytype=Integer)
     >>> d[42] = "galning"
+
+    INSTALLATION INSTRUCTIONS
+
+    I use debian, so I do the following:
+    $ sudo apt-get install python-setuptools python-mysqldb
+    $ easy_install sqlalchemy
+    $ easy_install sqldict
     """
 
     conn = None
@@ -732,13 +888,18 @@ def sqldict(db, tablename,
                     keytype=keytype,
                     valtype=valtype,
                     valgetter=valgetter,
+                    getrow=getrow,
                     keymod=keymod,
                     valmodin=valmodin,
                     valmodout=valmodout,
+                    query=query,
                     iterate=iterate,
+                    columns=columns,
+                    echo=echo,
                     )
 
 
 if __name__ == '__main__':
     import doctest
-    doctest.testmod()
+    doctest.testmod(globs=locals())
+
