@@ -3,6 +3,7 @@ from sqldict import sqldict, sqldict_old, _KEYSIZE, _VALSIZE
 import doctest
 import unittest
 from sqlalchemy import *
+from datetime import datetime
 
 SQLITEFS_FILE = "//tmp/.sqlitedb_test_sqlitefs"
 
@@ -32,6 +33,12 @@ class Test_sqldict_basic(unittest.TestCase):
                 xval varchar(%(valsize)s)
             )
         """,
+        t_char_int = """
+            CREATE TABLE t_char_int (
+                xkey varchar(%(valsize)s),
+                xval integer
+            )
+        """,
         t_int_char = """
             CREATE TABLE t_int_char (
                 xkey integer,
@@ -55,7 +62,10 @@ class Test_sqldict_basic(unittest.TestCase):
                         valsize = _VALSIZE,
             )
             for engine in self._engines.itervalues():
-                engine.execute(query)
+                try:
+                    engine.execute(query)
+                except:
+                    pass
 
     def tearDown(self):
         import os
@@ -101,6 +111,10 @@ class Test_sqldict_basic(unittest.TestCase):
             self.custom_valgetter(engine)
             self.keymod_valmod(engine)
             self.iterate(engine)
+            self.custom_query(engine)
+            self.setvalues(engine)
+            self.dictsetter(engine)
+            self.custom_columns(engine)
             factor = 1
             self.large_update(engine, count=factor*10)
 
@@ -190,6 +204,7 @@ class Test_sqldict_basic(unittest.TestCase):
             d.drop()
 
     def iterate(self, engine):
+        pass
         for val in ["asd", "qwe", "123", "haha", "hoho"]:
             engine.execute("""
                 INSERT INTO t_char_char
@@ -225,8 +240,81 @@ class Test_sqldict_basic(unittest.TestCase):
             if i==3: assert val == "hoho"
             if i==4: assert val == "qwe"
         assert i==4
+        #
+        # using iterate()
+        #
+        d.set(iterate=0)
+        for i, val in enumerate(d.iterate("foo")):
+            if i==0: assert val == "123"
+            if i==1: assert val == "asd"
+            if i==2: assert val == "haha"
+            if i==3: assert val == "hoho"
+            if i==4: assert val == "qwe"
+        assert i==4
 
 
+    def custom_query(self, engine):
+        for val in [1,2,3]:
+            engine.execute("""
+                INSERT INTO t_char_int
+                VALUES ('foo', %d)
+            """ % val )
+
+        for val in [2,3,4]:
+            engine.execute("""
+                INSERT INTO t_char_int
+                VALUES ('bar', %d)
+            """ % val )
+
+        d = sqldict(engine, "t_char_int", "xkey", "xval", query="""
+            SELECT SUM(xval)
+            FROM t_char_int
+            WHERE xkey='%s'
+        """)
+        assert d["foo"] == 6
+        assert d["bar"] == 9
+
+    def setvalues(self, engine):
+        d = sqldict(engine, "t_char_char", "xkey", "xval")
+        d["xa"] = "ya"
+        d.setvalues("xa", dict(xval="barbar"))
+        for val in d.iterate("xa"):
+            #print val
+            pass
+
+    def dictsetter(self, engine):
+        d = sqldict(engine, "t_char_char", "xkey", "xval", getrow=1)
+        d["xirb"] = dict(xval="pp")
+        for key, val in d.iteritems():
+            #print key, val
+            pass
+        for val in d.itervalues():
+            #print val
+            pass
+
+    def custom_columns(self, engine):
+        cols = [
+            Column("kk", String(200)),
+            Column("a", Integer),
+            Column("b", String(200)),
+            Column("c", DateTime),
+        ]
+        d = sqldict(engine, "add_columns", "kk",# "a_nonexistent",
+                    columns=cols, create=1, getrow=1)
+        try:
+            d["plupp"] = dict(a=123, b="asd", c=datetime.now())
+            assert len(d) == 1
+            for key, val in d.iteritems():
+                assert type(val.a) in (int, long)
+                assert type(val.b) in (str, unicode)
+                assert type(val.c) is type(datetime.now())
+            for val in d.itervalues():
+                assert type(val.a) in (int, long)
+                assert type(val.b) in (str, unicode)
+                assert type(val.c) is type(datetime.now())
+        finally:
+            d.drop()
+
 
 
 
@@ -463,7 +551,6 @@ class Test_sqldict_old(unittest.TestCase):
             self.populate_test_asdasd(conn)
             from functools import partial
 
-
             mydict = partial(sqldict_old, conn=conn, tablename="test_asdasd")
             name_by_ref = mydict(   valcol="name", keycol="ref")
             name_by_age = mydict(   valcol="name", keycol="age")
